<?php
/**
 * Copyright © 2015 FiloBlu. All rights reserved.
 */

namespace FiloBlu\Flow\Model;

use FiloBlu\Flow\Model\Channel\In\Images;
use Magento\Framework\App\ObjectManager;
use Magento\Store\Model\ScopeInterface;

/**
 * Class Processor
 * @package FiloBlu\Flow\Model
 */
class Processor
{
    /**
     * Max tries before mark the flow as faulty
     *
     * @var integer
     */
    const MAX_IMAGE_PROCESS_RETRY_COUNT = 3;

    /**
     * @var string
     */
    const CHANNEL_TYPE_IMAGE_CODE = 'image';

    /**
     * @var integer
     */
    const LOWER_PROCESS_ORDER = 1000;

    /**
     * @var integer
     */
    const LOWEST_PRIORITY_WEIGHT = 1000;

    /**
     * Processing order of known flows
     *
     * @var array
     */
    protected static $_processOrder = [
        'From\Product' => 0,
        'From\Eavs' => 1,
        'From\Stock' => 2,
        'From\Multistock' => 3,
        'From\Price' => 4,
        'From\Brotherhood' => 5,
        'Imagesflow' => 6
    ];

    /**
     * Waiting condition
     *
     * An inbound flow will not be executed if a condition will met
     * @var array
     */
    protected static $_waitingQueue = [

        /* Eavs will no processed if there is ite in processing */
        'From\Eavs' => [
            [
                'condition' => 'flow = ? AND status NOT IN (?)',
                'arguments' => [
                    'From\Product',
                    [
                        Inboundflow::STATUS_ERROR,
                        Inboundflow::STATUS_PROCESSED
                    ]
                ]
            ]
        ]
    ];

    /**
     * Priority weights
     *
     * @var array
     */
    protected static $_priorityWeights = [
        Inboundflow::PRIORITY_CRITICAL => 0,
        Inboundflow::PRIORITY_HIGH => 1,
        Inboundflow::PRIORITY_NORMAL => 2,
        Inboundflow::PRIORITY_LOW => 4
    ];

    /**
     * @var \Monolog\Logger
     */
    protected $_logger;

    /**
     * @var array
     */
    protected $_config = [];

    /**
     * @var \FiloBlu\Flow\Helper\Data
     */
    protected $_helperData;

    /**
     * @var Channel
     */
    protected $_channel;

    /**
     * @var Inboundflow
     */
    protected $_inboundflow;

    /**
     * @var Imagesflow
     */
    protected $_imageflow;

    /**
     * @var array
     */
    protected $_channels = [];

    /**
     * @var array
     */
    protected $_inboundItems = [];

    /**
     * @var \Magento\Framework\App\ResourceConnection
     */
    protected $_resourceConnection;

    /**
     * @var \FiloBlu\Flow\Helper\System
     */
    protected $_systemHelper;

    /**
     * @var ObjectManager
     */
    protected $_objectManager;

    /**
     * Default priorities to process
     *
     * @var array
     */
    protected $_priority = [
        Inboundflow::PRIORITY_HIGH,
        Inboundflow::PRIORITY_NORMAL,
        Inboundflow::PRIORITY_LOW
    ];

    /**
     * Processor constructor.
     * @param \Magento\Framework\Model\Context $context
     * @param \Magento\Framework\Registry $registry
     * @param \FiloBlu\Flow\Helper\Data $helperData
     * @param \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider
     * @param Channel $channel
     * @param Inboundflow $inboundflow
     * @param Imagesflow $imageFlow
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Framework\App\ResourceConnection $resourceConnection
     * @param \FiloBlu\Flow\Helper\System $systemHelper
     * @param array $data
     */
    public function __construct(
        \Magento\Framework\Model\Context $context,
        \Magento\Framework\Registry $registry,
        \FiloBlu\Flow\Helper\Data $helperData,
        \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider,
        \FiloBlu\Flow\Model\Channel $channel,
        \FiloBlu\Flow\Model\Inboundflow $inboundflow,
        \FiloBlu\Flow\Model\Imagesflow $imageFlow,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Framework\App\ResourceConnection $resourceConnection,
        \FiloBlu\Flow\Helper\System $systemHelper,
        array $data = []
    ) {
        $this->_logger = $loggerProvider->getLogger();
        $this->_channel = $channel;
        $this->_scopeConfig = $scopeConfig;
        $this->_helperData = $helperData;
        $this->_inboundflow = $inboundflow;
        $this->_imageflow = $imageFlow;
        $this->_systemHelper = $systemHelper;
        $this->_objectManager = ObjectManager::getInstance();
        $attributesToCopy = $this->_objectManager->create('FiloBlu\Flow\Model\System\ProductAttributes')
            ->getAttributesToCopy()
            ->getColumnValues('attribute_code');

        $registry->register('attribute_configurable_copy', $attributesToCopy);

        $this->_resourceConnection = $resourceConnection;

        $this->init();
    }

    /**
     *
     */
    private function init()
    {
        $this->initConfig();
        $this->setChannels();
        $this->setInboundflowItems();
    }

    /**
     *
     */
    public function initConfig()
    {
        $collection = $this->_channel->getCollection();

        foreach ($collection as $channel) {

            $config_data = json_decode($channel->getData('data'));

            $channel_name = strtolower($channel->getData('name'));
            $channel_model_class = $this->_helperData->getChannelModels($config_data->config->flow);
            $channel_type = $channel->getData('type');

            $this->_config['channels']->{$channel_type}[$channel_name] = $channel_model_class;
        }
    }

    /**
     * @return array
     */
    public function setInboundflowItems()
    {
        $this->_logger->info(__METHOD__);

        if (!$this->_inboundItems) {

            $collection = $this->_inboundflow->getCollection();

            foreach ($collection as $inbound) {
                $this->_inboundItems[] = $inbound->getData('name');
            }
        }

        return $this->_inboundItems;
    }

    /**
     * Process flows with priority set to critical
     *
     * @throws \Exception
     */
    public function processCritical()
    {
        $this->addPriority(Inboundflow::PRIORITY_CRITICAL);

        $this->fetchChannels();
        $this->receiveData();
        $this->parse();
        $this->process();
    }

    /**
     * Add priority to process.
     *
     * @see \FiloBlu\Flow\Model\Inboundflow
     * @param $priority
     * @throws \Exception
     */
    protected function addPriority($priority)
    {
        if (!Inboundflow::isValidPriority($priority)) {
            throw new \Exception("Invalid priority {$priority}");
        }

        $isCritcal = in_array(Inboundflow::PRIORITY_CRITICAL, $this->_priority);

        if ($isCritcal || $priority === Inboundflow::PRIORITY_CRITICAL) {
            $this->_priority = [Inboundflow::PRIORITY_CRITICAL];
            return;
        }

        if (in_array($priority, $this->_priority)) {
            return;
        }

        $this->_priority[] = $priority;
    }

    /**
     * Fetch form channels on the specified bus
     *
     * @param int $bus
     * @throws \Exception
     */
    public function fetchChannels($bus = 0)
    {
        $channels = $this->getChannels();

        if (!count($channels)) {
            throw new \Exception('No channel configured');
        }

        /** @var  \FiloBlu\Flow\Model\Channel\In\AbstractModel $channel */
        foreach ($channels as $name => $channel) {

            if ($bus && $channel->getConfig('bus') != $bus) {
                continue;
            }

            $list = $channel->rlist();

            if (count($list) && $list !== false) {
                $this->populateModel($channel, $name, $list);
            }
        }
    }

    /**
     * Get all channels from config
     *
     * @return array
     */
    public function getChannels()
    {
        return $this->_channels;
    }

    /**
     * @param string $type
     * @return $this
     */
    public function setChannels($type = 'in')
    {
        $channels = $this->_config['channels'];

        foreach ((Array)$channels->$type as $name => $channelModelClass) {

            $channelClass = '\FiloBlu\Flow\Model\Channel' . "\\" . ucfirst($type) . "\\" . ucfirst($channelModelClass);

            $model = false;

            try {
                $model = ObjectManager::getInstance()->create($channelClass);
                $model->load($name);
            } catch (\Exception $e) {
                $this->_logger->info($e->getMessage());
            }

            if ($model) {
                $this->_channels[$name] = $model;
            } else {
                $this->_logger->error("Error in channel {$type} {$channelModelClass} config. Channel skipped");
            }
        }

        return $this;
    }

    /**
     * TODO : fix better logic!
     *
     * @param $channel
     * @param $channel_name
     * @param $list
     */
    protected function populateModel($channel, $channelName, $list)
    {
        $channelType = $channel->getConfig('type');
        $channelFlow = $channel->getConfig('flow');
        $channelPriority = $channel->getConfig('priority');
        $lid = null;

        switch ($channelType) {

            case self::CHANNEL_TYPE_IMAGE_CODE:

                $model = $this->_objectManager->create(Inboundflow::class);
                $model->setName("images_" . date('Ymd_Hmi') . ".csv");
                $model->setFlow($channelFlow);
                $model->setType($channelType);
                $model->setChannel($channelName);
                $model->setRetry(0);
                $model->setStatus(Inboundflow::STATUS_FOUND);
                $model->setPriority($channelPriority);

                try {
                    $model->save();
                    $lid = $model->getId();
                } catch (\Exception $e) {
                    $this->_logger->info($e->getMessage());
                }

                if (!$lid) {
                    //handle error here
                    return false;
                }


                foreach ($list as $item) {
                    $model = $this->_objectManager->create(Inboundflow::class);

//                    $model->setId(null);
                    $model->setMetaFile($lid);
                    $model->setImageName($item);
                    $model->setImageStatus(Inboundflow::STATUS_FOUND);
                    $model->setData('image_queue_time', date("Y-m-d H:i:s"));
                    try {
                        $model->save();
                    } catch (\Exception $e) {
                        $this->_logger->info($e->getMessage());
                    }
                }
                break;

            default:

                foreach ($list as $item) {

                    // se il file è già presente nella tabella inbound, non deve essere reimportato
                    if (in_array($item, $this->_inboundItems)) {
                        continue;
                    }

                    $model = $this->_objectManager->create(Inboundflow::class);
                    $model->setName($item);
                    $model->setFlow($channelFlow);
                    $model->setType($channelType);
                    $model->setChannel($channelName);
                    $model->setRetry(0);
                    $model->setStatus(Inboundflow::STATUS_FOUND);
                    $model->setPriority($channelPriority);
                    try {
                        $model->save();
                    } catch (\Exception $e) {
                        $this->_logger->info($e->getMessage());
                    }
                }
        }
    }

    /**
     * Retrieve remote files
     * @throws \Exception
     */
    public function receiveData()
    {
        $flows = $this->getInboundFlowsToProcess(Inboundflow::STATUS_FOUND);

        /**
         * @var Inboundflow $flow
         */
        foreach ($flows as $flow) {

            try {

                if ($flow->getStatus() == Inboundflow::STATUS_RECEIVING) {
                    continue;
                }

                $flow->setStatus(Inboundflow::STATUS_RECEIVING)->save();

                $channelName = $flow->getChannel();
                $channel = $this->getChannel($channelName);

                if ($channel === null) {
                    $flow->setStatus(Inboundflow::STATUS_ERROR);
                    $flow->setLog("No channel found with name {$channelName}");
                    $flow->save();
                    continue;
                }

                if (!($channel instanceof \FiloBlu\Flow\Model\Channel\In\AbstractModel)) {
                    $message = "Error in channel {$channel->model} config during receiving";
                    $flow->setLog($message);
                    $flow->setStatus(Inboundflow::STATUS_ERROR)->save();
                    continue;
                }

                $received = $channel->receiveFile($flow);

                if ($received) {
                    $flow->setStatus(Inboundflow::STATUS_RECEIVED)->save();
                } else {
                    $flow->setLog("Error while receiving file");
                    $flow->setStatus(Inboundflow::STATUS_ERROR)->save();
                }

            } catch (\Exception $e) {
                $flow->setLog($e->getMessage());
                $flow->setStatus(Inboundflow::STATUS_ERROR)->save();
            }
        }
    }

    /**
     * Retrieve inbound flows to be processed sorted by process order and priority
     * filtered by status
     *
     * @param $status
     * @return \Magento\Framework\DataObject[]
     * @throws \Exception
     */
    public function getInboundFlowsToProcess($status)
    {
        if (!Inboundflow::isValidStatus($status)) {
            throw new \Exception("Invalid status {$status}");
        }

        $flows = $this->_inboundflow->getCollection()
            ->addFieldToFilter('status', $status)
            ->addFieldToFilter('priority', ['in' => $this->_priority])
            ->getItems();

        usort($flows, function ($a, $b) {

            $pwa = self::getPriorityWeight($a);
            $pwb = self::getPriorityWeight($b);

            $poa = self::getProcessOrder($a);
            $pob = self::getProcessOrder($b);

            if ($poa === $pob) {

                if ($pwa === $pwb) {
                    return 0;
                }

                return ($pwa < $pwb) ? -1 : 1;
            }

            return ($poa < $pob) ? -1 : 1;
        });

        return $flows;
    }

    /**
     * Get the priority weight for the given flow
     *
     * @param Inboundflow $flow
     * @return integer
     */
    protected static function getPriorityWeight(Inboundflow $flow)
    {
        $priority = $flow->getPriority();

        if (isset(self::$_priorityWeights[$priority])) {
            return self::$_priorityWeights[$priority];
        }

        return self::LOWEST_PRIORITY_WEIGHT;
    }

    /**
     * Get the process order for the given flow
     *
     * @param Inboundflow $flow
     * @return integer
     */
    protected static function getProcessOrder(Inboundflow $flow)
    {
        $type = $flow->getFlow();

        if (isset(self::$_processOrder[$type])) {
            return self::$_processOrder[$type];
        }

        return self::LOWER_PROCESS_ORDER;
    }

    /**
     * Get channel from name
     *
     * @param string $name
     * @return Channel|null
     */
    public function getChannel($name)
    {
        return (isset($this->_channels[$name])) ? $this->_channels[$name] : null;
    }

    /**
     * Parse
     * @throws \Exception
     */
    public function parse()
    {
        $flows = $this->getInboundFlowsToProcess(Inboundflow::STATUS_RECEIVED);

        foreach ($flows as $flow) {

            if ($flow->getStatus() == Inboundflow::STATUS_PARSING) {
                continue;
            }

            $flow->setStatus(Inboundflow::STATUS_PARSING);
            $flow->save();

            $channelName = $flow->getChannel();
            $channel = $this->getChannel($channelName);

            if ($channel === null) {
                $flow->setStatus(Inboundflow::STATUS_ERROR);
                $flow->setLog("No channel found with name {$channelName}");
                $flow->save();
                continue;
            }

            try {

                if ($channel->parse($flow)) {
                    $flow->setStatus(Inboundflow::STATUS_PARSED);
                    $flow->save();
                    continue;
                }

                $flow->setLog('Error while parsing');
                $flow->setStatus(Inboundflow::STATUS_ERROR);
                $flow->save();
            } catch (\Exception $e) {
                $flow->setLog($e->getMessage());
                $flow->setStatus(Inboundflow::STATUS_ERROR);
                $flow->save();
            }
        }
    }

    /**
     * @throws \Exception
     */
    public function process()
    {
        $flows = $this->getInboundFlowsToProcess(Inboundflow::STATUS_PARSED);

        $this->_systemHelper->dropTriggers();

        foreach ($flows as $flow) {

            if ($flow->getStatus() == Inboundflow::STATUS_PROCESSING) {
                continue;
            }

            if (!$this->canRun($flow)) {
                continue;
            }

            $flow->setStatus(Inboundflow::STATUS_PROCESSING);
            $flow->save();

            $this->_logger->info($flow->getFlow());
            $type = $flow->getType();

            try {

                switch ($type) {
                    case self::CHANNEL_TYPE_IMAGE_CODE:
                        $this->processImages($flow);
                        break;
                    default:
                        $this->processFile($flow);
                }

            } catch (\Exception $e) {
                $flow->setStatus(Inboundflow::STATUS_ERROR);
                $flow->setLog(__METHOD__ . ': ' . $e->getMessage() . "\n\n" . $e->getTraceAsString());
                $flow->save();
            }
        }

        if ($this->canEnableIndexers()) {
            $this->_systemHelper->enableIndexers();
        }
    }

    /**
     * @param Inboundflow $inboundFlow
     * @return bool
     */
    protected function canRun(Inboundflow $inboundFlow)
    {
        $flowType = $inboundFlow->getFlow();

        if (!array_key_exists($flowType, self::$_waitingQueue)) {
            return true;
        }

        $conditions = self::$_waitingQueue[$flowType];

        $connection = $this->_resourceConnection->getConnection();
        $table = $this->_inboundflow->getResource()->getMainTable();

        $sql = $connection->select()->from($table);
        $params = [];

        foreach ($conditions as $condition) {
            $sql->orWhere($condition['condition']);
            foreach ($condition['arguments'] as $param) {
                if (is_array($param)) {
                    $params[] = join(',', $param);
                } else {
                    $params[] = $param;
                }
            }
        }

        $stmt = $connection->query($sql, $params);

        if (count($stmt->fetchAll()) > 0) {
            return false;
        }

        return true;

    }

    /**
     * @param $flow
     * @return mixed
     */
    public function processImages($flow)
    {
        if ($flow->getStatus() == Inboundflow::STATUS_ERROR) {
            return;
        }

        /** @var \FiloBlu\Flow\Model\Channel\In\Images $model */
        $model = ObjectManager::getInstance()->create(Images::class);

        $result = $model->insertData($flow);

        if ($result) {
            $flow->setStatus(Inboundflow::STATUS_PROCESSED);
            $flow->save();
            return;
        }

        $total = $model->getImageToProcessCount();
        $errors = $model->getProcessedImageErrorCount();

        $retry = $flow->getRetry();
        $log = $flow->getLog() . "Process #{$retry}: Image to process {$total} ({$errors} fault)\n";

        $flow->setRetry($retry + 1);

        if ($retry >= self::MAX_IMAGE_PROCESS_RETRY_COUNT) {
            $flow->setLog("Unable to complete the flow after {$retry} tries. Look at flow images table for more details.\n\n" . $log);
            $flow->setStatus(Inboundflow::STATUS_ERROR);
        } else {
            $flow->setLog($log);
            $flow->setStatus(Inboundflow::STATUS_PARSED);
        }

        return $flow->save();
    }

    /**
     * @param $flow
     * @return bool
     */
    public function processFile($flow)
    {
        $model = ObjectManager::getInstance()->create('FiloBlu\Flow\Model\\' . $flow->getFlow());

        if (!$model) {
            return false;
        }

        try {
            $model->processFileRows($flow);
            $flow->setComment($model->getErrorMessage());
        } catch (\Exception $e) {
            $flow->setStatus(Inboundflow::STATUS_ERROR);
            $this->_logger->info($e->getMessage());
            $flow->setLog($e->getMessage())->save();
        }

        return $flow->setStatus(Inboundflow::STATUS_PROCESSED)->save();
    }

    /**
     * @return bool
     */
    public function canEnableIndexers()
    {
        $connection = $this->_resourceConnection->getConnection();
        $table = $this->_inboundflow->getResource()->getMainTable();

        $stmt = $connection->query("SELECT `id` FROM {$table} WHERE `status` = ? ", [Inboundflow::STATUS_PROCESSING]);

        if (count($stmt->fetchAll()) > 0) {
            return false;
        }

        return true;
    }

    /**
     * @throws \Exception
     */
    public function clean()
    {
        $days = (int)$this->_scopeConfig->getValue('filoblu_flow/flow_clean/flow_clean_days',
            ScopeInterface::SCOPE_STORE);

        if (!$days) {
            throw new \Exception("No days for clean specified");
        }

        $collection = $this->_inboundflow->getCollection();
        $connection = $collection->getResource()->getConnection();
        $table = $this->_inboundflow->getResource()->getMainTable();

        $flows = $collection->addFieldToFilter('status',
            [
                Inboundflow::STATUS_PROCESSED,
                Inboundflow::STATUS_ERROR
            ]
        )
            ->addFieldToFilter('receiving_time',
                ['lt' => new \Zend_Db_Expr("DATE_SUB(CURDATE(), INTERVAL {$days} DAY)")]
            )->getItems();

        foreach ($flows as $flow) {

            $type = $flow->getType();
            $channel = $flow->getChannel();

            try {

                switch ($type) {

                    case self::CHANNEL_TYPE_IMAGE_CODE:
                        break;

                    default:

                        $result = $connection->fetchRow(
                            "SELECT MAX(`id`) AS `id` FROM `{$table}` WHERE `channel` = ? ",
                            [$channel],
                            \Zend_Db::FETCH_OBJ);

                        if ($result->id != $flow->getId()) {
                            $this->cleanFile($flow);
                        }
                }

            } catch (\Exception $e) {
                $flow->setLog($e->getMessage());
                $flow->save();
            }
        }
    }

    /**
     * @param Inboundflow $flow
     * @return bool
     */
    public function cleanFile(Inboundflow $flow)
    {
        $model = ObjectManager::getInstance()->create('FiloBlu\Flow\Model\\' . $flow->getFlow());

        if (!$model) {
            return false;
        }

        try {
            $model->clean($flow);
            $flow->setStatus(Inboundflow::STATUS_CLEANED)->save();
        } catch (\Exception $e) {
            $this->_logger->info($e->getMessage());
            $flow->setLog($e->getMessage());
            $flow->save();
        }
    }

}
